/*
Copyright 2010-2012 Michael Braiman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package ui.auto.core.data;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


import org.openqa.selenium.WebElement;

import ui.auto.core.data.Data.GeneratorType;
import ui.auto.core.data.generators.AddressGenerator;
import ui.auto.core.data.generators.AlphaNumericGenerator;
import ui.auto.core.data.generators.CustomListGenerator;
import ui.auto.core.data.generators.DateGenerator;
import ui.auto.core.data.generators.HumanNameGenerator;
import ui.auto.core.data.generators.WordGenerator;
import ui.auto.core.data.generators.AddressGenerator.Address;
import ui.auto.core.pageobjects.WebComponent;

import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.annotations.XStreamAlias;
import com.thoughtworks.xstream.annotations.XStreamAsAttribute;
import com.thoughtworks.xstream.annotations.XStreamOmitField;


public class DataPersistence {
	@XStreamOmitField
	protected static XStream xstream;
	
	@XStreamAlias("xmlns")
	@XStreamAsAttribute
	protected String xmlns;
	
	@XStreamAlias("xmlns:xsi")
	@XStreamAsAttribute
	protected String xsi;
	
	@XStreamAlias("xsi:schemaLocation")
	@XStreamAsAttribute
	protected String schemaLocation;
	
	@XStreamOmitField
	private AddressGenerator addrGen=null;
	@XStreamOmitField
	private Address address = null;
	@XStreamOmitField
	private HumanNameGenerator nameGen=null;
	@XStreamOmitField
	private WordGenerator wordGen=null;
	@XStreamOmitField
	private Map<String,String> storage=new HashMap<String, String>();
	
	protected static void init(){
		xstream=new XStream();
		xstream.registerConverter(new WebComponentDataConverter());
	}
	
	
	public String toXML(){
		init();
		xstream.processAnnotations(this.getClass());
		String header="<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\n";
		return header + xstream.toXML(this);
	}
	
	public void toFile(String filePath){
		init();
		FileOutputStream fos=null;
		try {
			fos=new FileOutputStream(filePath);
		} catch (FileNotFoundException e) {
			throw new RuntimeException(e.getLocalizedMessage());
		}
		try {
			if (fos!=null) fos.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\n".getBytes());
			xstream.processAnnotations(this.getClass());
			xstream.toXML(this, fos);
			if (fos!=null) fos.close();
		} catch (IOException e) {
			throw new RuntimeException(e.getLocalizedMessage());
		}
	}
	
	@SuppressWarnings("unchecked")
	public static <T> T fromXml(String xml,Class<T> forClass){	
		init();
		xstream.processAnnotations(forClass);
		return (T) xstream.fromXML(xml);
	}
	
	@SuppressWarnings("unchecked")
	public static <T> T fromFile(String filePath,Class<T> forClass){
		init();
		T dataDef=null;
		FileInputStream fis=null;
		try {
			fis=new FileInputStream(filePath);
		} catch (FileNotFoundException e) {
			throw new RuntimeException(e.getLocalizedMessage());
		}
		try {
			xstream.processAnnotations(forClass);
			dataDef=(T) xstream.fromXML(fis);
			if (fis!=null) fis.close();
		} catch (Exception e) {
			throw new RuntimeException("Caused by " + e.getClass() + " : " + e.getLocalizedMessage());
		}
		return dataDef;
	}
	
	@SuppressWarnings("unchecked")
	public static <T> T deepCopy(T object){
		init();
		return (T) xstream.fromXML(xstream.toXML(object));
	}
	
	private String generate(Field field){
		String value=field.getAnnotation(Data.class).value();
		GeneratorType type=field.getAnnotation(Data.class).type();
		String init=field.getAnnotation(Data.class).init();
		String storageKey =field.getAnnotation(Data.class).storageKey();
		String returnValue="";
		
		if (storage.containsKey(storageKey) && value.isEmpty()) return storage.get(storageKey);
		switch (type){
			case FIXED_VALUE:
				returnValue=value;
				break;
			case ADDRESS:
				if (addrGen==null) addrGen=new AddressGenerator();
				if (!init.isEmpty() || address==null) {
					address=addrGen.generateAddress();
				}
				returnValue=address.toString(value);
				break;
			case CUSTOM_LIST:
				CustomListGenerator genList=new CustomListGenerator(value.split(","));
				returnValue=genList.getValue();
				break;
			case ALPHANUMERIC:
				AlphaNumericGenerator genAlpha=new AlphaNumericGenerator();
				returnValue= genAlpha.generate(value);
				break;
			case DATE:
				String[] args=init.split("\\|");
				try {
					DateGenerator dateGen=new DateGenerator(args[0],args[1],args[2]);
					if (value.isEmpty()){
						returnValue=dateGen.getDate();
					} else {
						returnValue=dateGen.getDate(value);
					}
					
				} catch (ParseException e) {
					throw new RuntimeException(e.getLocalizedMessage());
				}
				break;
			case HUMAN_NAMES:
				if (nameGen==null) nameGen=new HumanNameGenerator();
				returnValue=nameGen.getFullName(value);
				break;
			case WORD:
				if (wordGen==null) wordGen=new WordGenerator();
				returnValue=wordGen.generate(value);
				break;
			
		}
		if (!storageKey.isEmpty()) storage.put(storageKey, returnValue);
		return returnValue;
	}
	
	
	public void generateDataSet(Object dataSet) throws Exception{
		Field[] fields= dataSet.getClass().getDeclaredFields();
		for (Field field:fields){
			field.setAccessible(true);
			if (field.isAnnotationPresent(XStreamOmitField.class)) continue;
			String dataValue;
			if (field.isAnnotationPresent(Data.class)){
				dataValue=generate(field);
			} else {
				dataValue=field.getName();
			}
			if (field.getType().equals(String.class)){
				field.set(dataSet,dataValue);
			}else if (field.getType().equals(Boolean.class) || field.getType().equals(boolean.class)){
				field.set(dataSet, dataValue.equals("true"));
			} else if (field.getType().equals(Integer.class) || field.getType().equals(int.class)){
				field.set(dataSet,dataValue);
			} else if (WebComponent.class.isAssignableFrom(field.getType())){
				Object dataObj=field.getType().getConstructor(String.class).newInstance(dataValue);
				field.set(dataSet, dataObj);
			}else if (field.getType().equals(List.class)){
				String genericClz=field.getGenericType().toString().split("<")[1].replace(">","");
				List<Object> dataObj=new ArrayList<Object>();
				for (int i=0;i<3;i++){
					Object genObj=Class.forName(genericClz).newInstance();
					generateDataSet(genObj);
					dataObj.add(genObj);
				}
				field.set(dataSet, dataObj);
			}else if (field.getType().equals(WebElement.class)){
				//Skip, this field is not supporting WebComponentData interface
			} else {
				Object dataObj;
				try {
					dataObj=field.getType().newInstance();
				} catch (java.lang.InstantiationException e){
					String[] type=field.getType().getName().split("\\.");
					String[] cls=dataSet.getClass().getName().split("\\.");
					throw new RuntimeException("Can't instantiate '" + type[type.length-1] + "' field named '" + field.getName() + "' in '" + cls[cls.length-1]+"'");
				}
				field.set(dataSet, dataObj);
				generateDataSet(dataObj);
			}
		}
	}
	
	public String generateXML() throws Exception{
		Object dataSet=this.getClass().newInstance();
		generateDataSet(dataSet);
		return ((DataPersistence)dataSet).toXML();
	}
	
}
